#include	<stdio.h>
#include	<stdlib.h>
#include	<signal.h>
#include	<errno.h>

#include	<unistd.h>
#include	<sys/types.h>
#include	<sys/wait.h>

#include	<X11/X.h>
#include	<X11/Xos.h>
#include	<X11/Xlib.h>
#include	<X11/Xutil.h>
#include	<X11/Xatom.h>

#include	"wm.h"

Mode mode;						/* The window manager's mode. (See "wm.h".) */
int start_x;					/* The X position where the mode changed. */
int start_y;					/* The Y position where the mode changed. */

Display *dpy;					/* The connection to the X server. */
int display_width;			/* The width of the screen we're managing. */
int display_height;			/* The height of the screen we're managing. */

Window root;					/* Root window of managed screen. */
Window popup;					/* Third-button menu. */

GC gc;							/* The default GC. */
GC gc_thin;						/* The default GC but with thinner lines. */
GC popup_gc;					/* The GC for the popup window (menu). */

unsigned long black;			/* Black pixel. */
unsigned long white;			/* White pixel. */

XFontStruct *font;			/* Actual titlebar font. */
XFontStruct *popup_font;	/* Actual menu font. */

Bool shape;						/* Does server support Shape Window extension? */
int shape_event;				/* ShapeEvent event type. */

/* Atoms we're interested in. See the ICCCM for more information. */
Atom wm_state;
Atom wm_change_state;
Atom wm_protocols;
Atom wm_delete;
Atom wm_take_focus;
Atom wm_colormaps;

static void scan_window_tree(void);

extern int
main(void)
{
	XEvent ev;
	XGCValues gv;
	XSetWindowAttributes attr;

	mode = wm_initialising;

	/* Open a connection to the X server. */
	dpy = XOpenDisplay("");
	if (dpy == 0)
		Panic("can't open display.");

	get_resources();

	/* Find the root window. */
	root = DefaultRootWindow(dpy);
	display_width = DisplayWidth(dpy, DefaultScreen(dpy));
	display_height = DisplayHeight(dpy, DefaultScreen(dpy));

	/* Set up an error handler. */
	XSetErrorHandler(ErrorHandler);

	/* Set up signal handlers. */
	signal(SIGTERM, Terminate);
	signal(SIGINT, Terminate);
	signal(SIGHUP, Terminate);

	/* Internalize useful atoms. */
	wm_state = XInternAtom(dpy, "WM_STATE", False);
	wm_change_state = XInternAtom(dpy, "WM_CHANGE_STATE", False);
	wm_protocols = XInternAtom(dpy, "WM_PROTOCOLS", False);
	wm_delete = XInternAtom(dpy, "WM_DELETE_WINDOW", False);
	wm_take_focus = XInternAtom(dpy, "WM_TAKE_FOCUS", False);
	wm_colormaps = XInternAtom(dpy, "WM_COLORMAP_WINDOWS", False);

	/* Get the pixel values of the only two colours we use. */
	black = BlackPixel(dpy, DefaultScreen(dpy));
	white = WhitePixel(dpy, DefaultScreen(dpy));

	/*
	 *	Get fonts for our titlebar and our popup window. We try to
	 * get Lucida, but if we can't we make do with fixed because everyone
	 * has that.
	 */
	font = XLoadQueryFont(dpy, font_name);
	if (font == 0)
		font = XLoadQueryFont(dpy, "fixed");
	if (font == 0)
		Panic("can't find a font for the titlebars.");

	popup_font = XLoadQueryFont(dpy, popup_font_name);
	if (popup_font == 0)
		popup_font = XLoadQueryFont(dpy, "fixed");
	if (popup_font == 0)
		Panic("can't find a font for the popup window.");

	/* Set up our mouse cursors. */
	Cursor_Initialise();

	/* See if the server has the Shape Window extension. */
	shape = servershapes();

	/* Set up root (frame) GC's. */
	gv.foreground = black ^ white;
	gv.background = white;
	gv.font = font->fid;
	gv.function = GXxor;
	gv.line_width = 1;
	gv.subwindow_mode = IncludeInferiors;
	gc_thin = XCreateGC(dpy, root, GCForeground | GCBackground |
	    GCFunction | GCFont | GCLineWidth | GCSubwindowMode, &gv);

	gv.line_width = 2;
	gc = XCreateGC(dpy, root, GCForeground | GCBackground | GCFunction |
	    GCFont | GCLineWidth | GCSubwindowMode, &gv);

	/* Create a window for our popup. */
	popup = XCreateSimpleWindow(dpy, root, 0, 0, 1, 1, 1, black, white);
	attr.cursor = frame_cursor;
	attr.event_mask = ButtonMask | ButtonMotionMask | ExposureMask;
	XChangeWindowAttributes(dpy, popup, CWCursor | CWEventMask, &attr);

	/* Create popup window's GC. */
	gv.line_width = 1;
	gv.font = popup_font->fid;
	popup_gc = XCreateGC(dpy, popup, GCForeground | GCBackground |
	    GCFunction | GCFont | GCLineWidth | GCSubwindowMode, &gv);

	/* Announce our interest in the root window. */
	attr.cursor = root_cursor;
	attr.event_mask = SubstructureRedirectMask
	    | SubstructureNotifyMask | ColormapChangeMask
	    | ButtonPressMask | PropertyChangeMask | EnterWindowMask;
	XChangeWindowAttributes(dpy, root, CWCursor | CWEventMask, &attr);

	/* Make sure all our communication to the server got through. */
	XSync(dpy, False);

	/* See who needs looking after. */
	scan_window_tree();

	/*
	 * Initialisation is finished, but we start off not interacting with the
	 * user.
	 */
	mode = wm_idle;

	/*
	 * The main event loop.
	 */
	for (;;) {
		XNextEvent(dpy, &ev);
		dispatch(&ev);
	}
}

void
SendConfigureNotify(Client *c)
{
	XConfigureEvent ce;

	ce.type = ConfigureNotify;
	ce.event = c->window;
	ce.window = c->window;
	ce.x = c->size.x + border;
	ce.y = c->size.y + border;
	ce.width = c->size.width - 2 * border;
	ce.height = c->size.height - 2 * border;
	ce.above = None;
	ce.border_width = c->border;
	ce.override_redirect = 0;
	XSendEvent(dpy, c->window, False, StructureNotifyMask, (XEvent *) & ce);
}

static void
scan_window_tree(void)
{
	unsigned int	i;
	unsigned int	nwins;
	Client	*c;
	Window	dw1;
	Window	dw2;
	Window	*wins;
	XWindowAttributes	attr;

	XQueryTree(dpy, root, &dw1, &dw2, &wins, &nwins);
	for (i = 0; i < nwins; i++) {
		XGetWindowAttributes(dpy, wins[i], &attr);
		if (attr.override_redirect || isshaped(wins[i]) || wins[i] == popup)
			continue;
		c = Client_Add(wins[i]);
		if (c != 0 && c->window == wins[i]) {
			c->size.x = attr.x - border;
			c->size.y = attr.y - border;
			c->size.width  = attr.width  + 2 * border;
			c->size.height = attr.height + 2 * border;
			c->border = attr.border_width;
			if (attr.map_state == IsViewable) {
				c->internal_state = IPendingReparenting;
				manage(c, 1);
			}
		}
	}
	XFree(wins);
}

extern void
shell(void)
{
	char *shell = getenv("SHELL");

	if (shell == 0)
		shell = "/bin/sh";

	if (fork() == 0) {
		/* Start a shell as a child, then exit. */
		if (fork() == 0) {
			close(ConnectionNumber(dpy));

			execl(shell, shell, "-c", terminal, 0);
			fprintf(stderr, "lwm: can't exec \"%s -c %s\"\n", shell, terminal);
			/* FALLTHROUGH */
			execlp("xterm", "xterm", 0);
			exit(EXIT_FAILURE);
		}

		/* Exit, shell becomes child of init. */
		exit(EXIT_SUCCESS);
	}

	/* Wait for child. */
	wait(0);
}

extern int
titleheight(void)
{
	return font->ascent + font->descent + 1;
}
